// Copyright (C) 2018-2019 Intel Corporation
// SPDX-License-Identifier: Apache-2.0
//

/**
* \brief The entry point for the Inference Engine object_detection demo application
* \file object_detection_demo_ssd_async/main.cpp
* \example object_detection_demo_ssd_async/main.cpp
*/

#include <gflags/gflags.h>

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/utsname.h>
#include <linux/i2c-dev.h>
#include <linux/i2c.h>
#include <sys/ioctl.h>
#include <queue>

#include <chrono>
#include <iostream>
#include <memory>
#include <vector>
#include <string>
#include <algorithm>

#include <inference_engine.hpp>
#include <ngraph/ngraph.hpp>

#include <monitors/presenter.h>
#include <samples/ocv_common.hpp>
#include <samples/slog.hpp>

#include "object_detection_demo_ssd_async_cam_scaler_multi.hpp"

#define VFB_ADDR 0x7F000000
#define VFB_SIZE 0x00800000
#define LW_BRIDGE_ADDR 0x0
#define LW_BRIDGE_SIZE 0x1000

/* VFB II control registers */
#define VFB_OFFSET				0x000
#define ALTVIPFB2_CONTROL		0
#define ALTVIPFB2_STATUS		1
#define ALTVIPFB2_INTERRUPT		2
#define ALTVIPFB2_FRAME_COUNTER	3
#define ALTVIPFB2_FRAME_DROP	4
#define ALTVIPFB2_FRAME_INFO	5
#define ALTVIPFB2_FRAME_START	6
#define ALTVIPFB2_FRAME_READER	7
#define ALTVIPFB2_FRAME_MISC	8
#define ALTVIPFB2_FRAME_LOCK	9
#define ALTVIPFB2_FRAME_IRATE	10
#define ALTVIPFB2_FRAME_ORATE	11

/* GPIO registers */
#define PIO_OFFSET				0x040
#define ALTPIO_DATA				0
//	[9:8]	raw alignment	= b00:	R  Gr Gb B
//							  b01:	Gr R  B  Gb
//							  b10:	Gb B  R  Gr
//							  b11:	B  Gb Gr R
//	[7:6]	8bit select		= b11:  [9:2]
//							  b10:  [8:1]
//							  else: [7:0]
//	[5:4]	size select		= b00:  1920x1080(FHD)
//							  b11:  640x480(VGA)
//	[3]		mipi0_gpio[0]
//	[2]		mipi0_gpio[1]
//	[1]		cam_nrst
//	[0]		not use

/* Scaler II control registers */
#define SCL_OFFSET				0x200
#define ALTVIPSCL2_CONTROL		0
#define ALTVIPSCL2_STATUS		1
#define ALTVIPSCL2_INTERRUPT	2
#define ALTVIPSCL2_WIDTH		3
#define ALTVIPSCL2_HEIGHT		4

/* Chip ID registers */
#define	ID_OV5647					0x5647
#define	ID_TC358748					0x4401
#define	MAX_RETRY_CAM_ON			10
#define	WAIT_RETRY_TIME				1000000

/* Maximum VPU count */
#define MAX_VPU 				4
/* Timer average tables */
#define TIME_TABLES				20

void		cis_init		( int, int );
void		m2p_init		( void );
void		cis_start		( void );
void		cis_stop		( void );

using namespace InferenceEngine;

bool ParseAndCheckCommandLine(int argc, char *argv[]) {
    // ---------------------------Parsing and validation of input args--------------------------------------
    gflags::ParseCommandLineNonHelpFlags(&argc, &argv, true);
    if (FLAGS_h) {
       showUsage();
       showAvailableDevices();
       return false;
    }
    slog::info << "Parsing input parameters" << slog::endl;

    if (FLAGS_i.empty()) {
        throw std::logic_error("Parameter -i is not set");
    }

    if (FLAGS_m.empty()) {
        throw std::logic_error("Parameter -m is not set");
    }

    return true;
}

void frameToBlob(const cv::Mat& frame,
                 InferRequest::Ptr& inferRequest,
                 const std::string& inputName) {
    if (FLAGS_auto_resize) {
        /* Just set input blob containing read image. Resize and layout conversion will be done automatically */
        inferRequest->SetBlob(inputName, wrapMat2Blob(frame));
    } else {
        /* Resize and copy data from the image to the input blob */
        Blob::Ptr frameBlob = inferRequest->GetBlob(inputName);
        matU8ToBlob<uint8_t>(frame, frameBlob);
    }
}

int main(int argc, char *argv[]) {
	int				result = 0;			// Result value
	unsigned char	*frame_buf = NULL;	// Local frame memory
	int          	fd_vfb = 0;			// uio0 file discripter
	unsigned char	*frame_mem = NULL;	// uio0 mmap pointer
	int 			fd_reg = 0;			// uio1 file discripter
	unsigned char	*reg_mem = NULL;	// uio1 mmap pointer
	unsigned long	*vfb_reg = NULL;	// Frame buffer register base pointer
	unsigned long	*scl_reg = NULL;	// Scaler register base pointer
	unsigned long	*pio_reg = NULL;	// Camera PIO register base pointer
	bool			fpd_input = false;	// Input type
	int				video_res = 1080;
	int				data_mode = 10;

    try {
        /** This demo covers certain topology and cannot be generalized for any object detection **/
        std::cout << "InferenceEngine: " << GetInferenceEngineVersion() << std::endl;

        // ------------------------------ Parsing and validation of input args ---------------------------------
        if (!ParseAndCheckCommandLine(argc, argv)) {
            return 0;
        }
        slog::info << "Reading input" << slog::endl;
        slog::info << "OpenCV version "<< CV_VERSION << slog::endl;

		int width  = FLAGS_iw;
		int height = FLAGS_ih;
		cv::VideoCapture 	cap;
	    cv::Mat 			curr_frame;
	    cv::Mat 			next_frame;
		fpd_input = (FLAGS_i == "FPD")? true : false;
		float font_size = (float)width / 640.0 * 0.6;

		slog::info << "Capture size: " << width << " x "<< height << slog::endl;

		if(!fpd_input) {
		    if (!((FLAGS_i == "cam") ? cap.open(0) : cap.open(FLAGS_i.c_str()))) {
		        throw std::logic_error("Cannot open input file or camera: " + FLAGS_i);
		    }

			slog::info << "cap.open success" << slog::endl;
			cap.set(cv::CAP_PROP_FOURCC, cv::VideoWriter::fourcc('M','J','P','G'));
		    cap.set(cv::CAP_PROP_FRAME_WIDTH, width);
		    cap.set(cv::CAP_PROP_FRAME_HEIGHT, height);

			slog::info << "cap.get success" << slog::endl;

		    // read input (video) frame
		    cap >> curr_frame;

		    if (!cap.grab()) {
		        throw std::logic_error("This demo supports only video (or camera) inputs !!! "
		                               "Failed getting next frame from the " + FLAGS_i);
		    }
		} else {
			if ((fd_vfb = open("/dev/uio0", O_RDWR)) != -1) 
			{
				frame_mem = (unsigned char *)mmap(NULL, VFB_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd_vfb, 0);
				if(frame_mem == (unsigned char *)-1)
				{
					throw std::logic_error("/dev/uio0 mmap error!");
				}			
			} else
			{
				throw std::logic_error("/dev/uio0 open error!");
			}

			if ((fd_reg = open("/dev/uio1", O_RDWR)) != -1) 
			{
				reg_mem = (unsigned char *)mmap(NULL, LW_BRIDGE_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd_reg, LW_BRIDGE_ADDR);
				if(reg_mem == (unsigned char *)-1)
				{
					throw std::logic_error("/dev/uio1 mmap error!");
				}			
			} else
			{
				throw std::logic_error("/dev/uio0 open error!");
			}
			
			pio_reg							= (unsigned long*)( reg_mem + PIO_OFFSET );
			pio_reg[ALTPIO_DATA]			= 0x0000;			// Reset camera input path
			usleep( 100 );
	
			// Activation camera input path
			if		( video_res == 1080 )	pio_reg[ALTPIO_DATA]	= 0x030e;		// 1920x1080(FHD)
			else if	( video_res == 480 )	pio_reg[ALTPIO_DATA]	= 0x033e;		//  640x480 (VGA)

			usleep( 10000 );
			// Initialize camera input path
			m2p_init();
			cis_init( video_res, data_mode );
			usleep( 100 );
	        slog::info << "OV5647 & TC358748 initialize done" << slog::endl;

			vfb_reg = (unsigned long*)(reg_mem + VFB_OFFSET);
			scl_reg = (unsigned long*)(reg_mem + SCL_OFFSET);
			// Initialize Scaler
			scl_reg[ALTVIPSCL2_WIDTH] = width;
			scl_reg[ALTVIPSCL2_HEIGHT] = height;
			scl_reg[ALTVIPSCL2_CONTROL] = 0x01;	// Scaler Start

			// Initialize Frame Buffer
			vfb_reg[ALTVIPFB2_CONTROL] = 0x01;  // Frame buffer Start

			// Start Camera input
			cis_start();
	        slog::info << "Scaler & Frame buffer Start" << slog::endl;
			usleep( 10000 );

			// Enable camera input into FPGA
			if		( video_res == 1080 )	pio_reg[ALTPIO_DATA]	= 0x030f;		// 1920x1080(FHD)
			else if	( video_res == 480 )	pio_reg[ALTPIO_DATA]	= 0x033f;		//  640x480 (VGA)

		    // read input (video) frame
			if((frame_buf = (unsigned char *)malloc(width * height * 3)) == NULL)
			{
				throw std::logic_error("Loca frame buffer can not get!");
			}
			unsigned char *frame_adr = frame_mem + (vfb_reg[ALTVIPFB2_FRAME_START] - VFB_ADDR);
			memcpy(frame_buf, frame_adr, width * height * 3);
		    curr_frame = cv::Mat(cv::Size(width, height), CV_8UC3, frame_buf);
		}
        // -----------------------------------------------------------------------------------------------------

        // --------------------------- 1. Load inference engine -------------------------------------
        slog::info << "Loading Inference Engine" << slog::endl;
        Core ie;

        slog::info << "Device info: " << slog::endl;
        std::cout << ie.GetVersions(FLAGS_d);

        /** Load extensions for the plugin **/

        if (!FLAGS_l.empty()) {
            // CPU(MKLDNN) extensions are loaded as a shared library and passed as a pointer to base extension
            IExtensionPtr extension_ptr = make_so_pointer<IExtension>(FLAGS_l.c_str());
            ie.AddExtension(extension_ptr, "CPU");
        }
        if (!FLAGS_c.empty()) {
            // clDNN Extensions are loaded from an .xml description and OpenCL kernel files
            ie.SetConfig({{PluginConfigParams::KEY_CONFIG_FILE, FLAGS_c}}, "GPU");
        }

        /** Per layer metrics **/
        if (FLAGS_pc) {
            ie.SetConfig({ { PluginConfigParams::KEY_PERF_COUNT, PluginConfigParams::YES } });
        }
        // -----------------------------------------------------------------------------------------------------

        // --------------------------- 2. Read IR Generated by ModelOptimizer (.xml and .bin files) ------------
        slog::info << "Loading network files" << slog::endl;
        /** Read network model **/
        auto cnnNetwork = ie.ReadNetwork(FLAGS_m);
        /** Set batch size to 1 **/
        slog::info << "Batch size is forced to  1." << slog::endl;
        cnnNetwork.setBatchSize(1);
        /** Read labels (if any)**/
        std::string labelFileName = fileNameNoExt(FLAGS_m) + ".labels";
        std::vector<std::string> labels;
        std::ifstream inputFile(labelFileName);
        std::copy(std::istream_iterator<std::string>(inputFile),
                  std::istream_iterator<std::string>(),
                  std::back_inserter(labels));
        // -----------------------------------------------------------------------------------------------------

        /** SSD-based network should have one input and one output **/
        // --------------------------- 3. Configure input & output ---------------------------------------------
        // --------------------------- Prepare input blobs -----------------------------------------------------
        slog::info << "Checking that the inputs are as the demo expects" << slog::endl;
        InputsDataMap inputInfo(cnnNetwork.getInputsInfo());

        std::string imageInputName, imageInfoInputName;
        size_t netInputHeight, netInputWidth;

        for (const auto & inputInfoItem : inputInfo) 
		{
            if (inputInfoItem.second->getTensorDesc().getDims().size() == 4) 
			{  // first input contains images
                imageInputName = inputInfoItem.first;
                inputInfoItem.second->setPrecision(Precision::U8);
                if (FLAGS_auto_resize) 
				{
                    inputInfoItem.second->getPreProcess().setResizeAlgorithm(ResizeAlgorithm::RESIZE_BILINEAR);
                    inputInfoItem.second->getInputData()->setLayout(Layout::NHWC);
                } else 
				{
                    inputInfoItem.second->getInputData()->setLayout(Layout::NCHW);
                }
                const TensorDesc& inputDesc = inputInfoItem.second->getTensorDesc();
                netInputHeight = getTensorHeight(inputDesc);
                netInputWidth = getTensorWidth(inputDesc);
            } else if (inputInfoItem.second->getTensorDesc().getDims().size() == 2) 
			{  // second input contains image info
                imageInfoInputName = inputInfoItem.first;
                inputInfoItem.second->setPrecision(Precision::FP32);
            } else 
			{
                throw std::logic_error("Unsupported " +
                                       std::to_string(inputInfoItem.second->getTensorDesc().getDims().size()) + "D "
                                       "input layer '" + inputInfoItem.first + "'. "
                                       "Only 2D and 4D input layers are supported");
            }
        }

        // --------------------------- Prepare output blobs -----------------------------------------------------
        slog::info << "Checking that the outputs are as the demo expects" << slog::endl;
        OutputsDataMap outputInfo(cnnNetwork.getOutputsInfo());
        if (outputInfo.size() != 1) {
            throw std::logic_error("This demo accepts networks having only one output");
        }
        DataPtr& output = outputInfo.begin()->second;
        auto outputName = outputInfo.begin()->first;

		int num_classes = 0;
        if (auto ngraphFunction = cnnNetwork.getFunction()) {
            for (const auto op : ngraphFunction->get_ops()) {
                if (op->get_friendly_name() == outputName) {
                    auto detOutput = std::dynamic_pointer_cast<ngraph::op::DetectionOutput>(op);
                    if (!detOutput) {
                        THROW_IE_EXCEPTION << "Object Detection network output layer(" + op->get_friendly_name() +
                            ") should be DetectionOutput, but was " +  op->get_type_info().name;
                    }

                    num_classes = detOutput->get_attrs().num_classes;
                    break;
                }
            }
        } else {
            const CNNLayerPtr outputLayer = cnnNetwork.getLayerByName(outputName.c_str());
            if (outputLayer->type != "DetectionOutput") {
                throw std::logic_error("Object Detection network output layer(" + outputLayer->name +
                                       ") should be DetectionOutput, but was " +  outputLayer->type);
            }

            num_classes = outputLayer->GetParamAsInt("num_classes");
        }
        if (static_cast<int>(labels.size()) != num_classes) {
            if (static_cast<int>(labels.size()) == (num_classes - 1))  // if network assumes default "background" class, having no label
                labels.insert(labels.begin(), "fake");
            else
                labels.clear();
        }
        const SizeVector outputDims = output->getTensorDesc().getDims();
        const int maxProposalCount = outputDims[2];
        const int objectSize = outputDims[3];
        if (objectSize != 7) {
            throw std::logic_error("Output should have 7 as a last dimension");
        }
        if (outputDims.size() != 4) {
            throw std::logic_error("Incorrect output dimensions for SSD");
        }
        output->setPrecision(Precision::FP32);
        output->setLayout(Layout::NCHW);
        // -----------------------------------------------------------------------------------------------------

        // --------------------------- 4. Loading model to the device ------------------------------------------
        slog::info << "Loading model to the device" << slog::endl;
        std::string devices[MAX_VPU];
        std::vector<std::string> myriadDevices = ie.GetMetric("MYRIAD", METRIC_KEY(AVAILABLE_DEVICES));
        int vpu_count = (int)myriadDevices.size();
        vpu_count = (vpu_count > MAX_VPU)? MAX_VPU : (vpu_count > FLAGS_v)? FLAGS_v : vpu_count;
        ExecutableNetwork network[MAX_VPU];
        for (int i = 0; i < vpu_count ; i ++) 
        {
            devices[i] = std::string("MYRIAD.") + myriadDevices[i];
            slog::info << devices[i] << slog::endl;
            network[i] = ie.LoadNetwork(cnnNetwork, devices[i], {});
        }
        // -----------------------------------------------------------------------------------------------------

        // --------------------------- 5. Create infer request -------------------------------------------------
        slog::info << "Create infer request" << slog::endl;
        InferRequest::Ptr async_infer_request_curr[MAX_VPU];
        InferRequest::Ptr async_infer_request_next[MAX_VPU];
        for (int i = 0; i < vpu_count ; i ++) 
        {
		    async_infer_request_curr[i] = network[i].CreateInferRequestPtr();
		    async_infer_request_next[i] = network[i].CreateInferRequestPtr();
        }

        /* it's enough just to set image info input (if used in the model) only once */
        if (!imageInfoInputName.empty()) {
            auto setImgInfoBlob = [&](const InferRequest::Ptr &inferReq) {
                auto blob = inferReq->GetBlob(imageInfoInputName);
                auto data = blob->buffer().as<PrecisionTrait<Precision::FP32>::value_type *>();
                data[0] = static_cast<float>(netInputHeight);  // height
                data[1] = static_cast<float>(netInputWidth);  // width
                data[2] = 1;
            };
		    for (int i = 0; i < vpu_count ; i ++) 
		    {
		        setImgInfoBlob(async_infer_request_curr[i]);
		        setImgInfoBlob(async_infer_request_next[i]);
			}
        }
        // -----------------------------------------------------------------------------------------------------

        // --------------------------- 6. Do inference ---------------------------------------------------------
        slog::info << "Start inference " << slog::endl;

        bool isLastFrame = false;
        bool isAsyncMode = false;  // execution is always started using SYNC mode
        bool isModeChanged = false;  // set to TRUE when execution mode is changed (SYNC<->ASYNC)

        typedef std::chrono::duration<double, std::ratio<1, 1000>> ms;
        auto total_t0 = std::chrono::high_resolution_clock::now();
        auto wallclock = std::chrono::high_resolution_clock::now();
        double ocv_render_time = 0;

        std::cout << "To close the application, press 'CTRL+C' here or switch to the output window and press ESC key" << std::endl;
        std::cout << "To switch between sync/async modes, press TAB key in the output window" << std::endl;
        cv::Size graphSize{static_cast<int>(((fpd_input)? width : cap.get(cv::CAP_PROP_FRAME_WIDTH)) / 4), 60};
        Presenter presenter(FLAGS_u, static_cast<int>(((fpd_input)? height : cap.get(cv::CAP_PROP_FRAME_HEIGHT))) - graphSize.height - 10, graphSize);

		int table_ptr = 0;
		int table_cnt = 0;
		double	cap_table[TIME_TABLES];
		double	wal_table[TIME_TABLES];
		double	det_table[TIME_TABLES];

		int cur_idx = 0;
		int prv_idx = -1;
		for(cur_idx = 0;;cur_idx = (cur_idx >= (vpu_count - 1))? 0 : (cur_idx + 1)) {
            auto t0 = std::chrono::high_resolution_clock::now();
            // Here is the first asynchronous point:
            // in the async mode we capture frame to populate the NEXT infer request
            // in the regular mode we capture frame to the CURRENT infer request
			if(!fpd_input) {
		        if (!cap.read(next_frame)) {
		            if (next_frame.empty()) {
		                isLastFrame = true;  // end of video file
		            } else {
		                throw std::logic_error("Failed to get frame from cv::VideoCapture");
		            }
		        }
			} else {
				if(vfb_reg[ALTVIPFB2_FRAME_INFO] & 0x80000000)
				{
					unsigned char *frame_adr = frame_mem + (vfb_reg[ALTVIPFB2_FRAME_START] - VFB_ADDR);
					memcpy(frame_buf, frame_adr, width * height * 3);
				}
				next_frame = cv::Mat(cv::Size(width, height), CV_8UC3, frame_buf);
				vfb_reg[ALTVIPFB2_FRAME_MISC] = 0x1;
			}

            if (isAsyncMode) {
                if (isModeChanged) {
                    frameToBlob(curr_frame, async_infer_request_curr[cur_idx], imageInputName);
                }
                if (!isLastFrame) {
                    frameToBlob(next_frame, async_infer_request_next[cur_idx], imageInputName);
                }
            } else if (!isModeChanged) {
                frameToBlob(curr_frame, async_infer_request_curr[cur_idx], imageInputName);
            }

            auto t1 = std::chrono::high_resolution_clock::now();
            double ocv_decode_time = std::chrono::duration_cast<ms>(t1 - t0).count();

            t0 = std::chrono::high_resolution_clock::now();
            // Main sync point:
            // in the truly Async mode we start the NEXT infer request, while waiting for the CURRENT to complete
            // in the regular mode we start the CURRENT request and immediately wait for it's completion
            if (isAsyncMode) {
				prv_idx = cur_idx;
                if (isModeChanged) {
                    async_infer_request_curr[cur_idx]->StartAsync();
                }
                if (!isLastFrame) {
                    async_infer_request_next[cur_idx]->StartAsync();
                }
            } else if (!isModeChanged) {
                async_infer_request_curr[cur_idx]->StartAsync();
            }

            if (OK == async_infer_request_curr[(prv_idx < 0)? cur_idx : prv_idx]->Wait(IInferRequest::WaitMode::RESULT_READY)) {
                t1 = std::chrono::high_resolution_clock::now();
                ms detection = std::chrono::duration_cast<ms>(t1 - t0);

                t0 = std::chrono::high_resolution_clock::now();
                ms wall = std::chrono::duration_cast<ms>(t0 - wallclock);
                wallclock = t0;

                t0 = std::chrono::high_resolution_clock::now();

                presenter.drawGraphs(curr_frame);

				cap_table[table_ptr] = (ocv_decode_time + ocv_render_time);
				wal_table[table_ptr] = wall.count();
				det_table[table_ptr] = detection.count();
				table_ptr = (table_ptr >= (TIME_TABLES - 1))? 0 : table_ptr + 1;
				table_cnt = (table_cnt >= TIME_TABLES)? TIME_TABLES : table_cnt + 1;

				double cap_time = 0;
				double wal_time = 0;
				double det_time = 0;
				for(int i = 0;i < table_cnt;i ++)
				{
					cap_time += cap_table[i];				
					wal_time += wal_table[i];				
					det_time += det_table[i];				
				}
				cap_time /= table_cnt;				
				wal_time /= table_cnt;				
				det_time /= table_cnt;				

	            std::ostringstream out;
	            out << "OpenCV cap/render time: " << std::fixed << std::setprecision(2) << cap_time << " ms";
	            cv::putText(curr_frame, out.str(), cv::Point2f(0, 25), cv::FONT_HERSHEY_TRIPLEX, font_size, cv::Scalar(0, 255, 0));
	            out.str("");
	            out << "Wallclock time " << (isAsyncMode ? "(TRUE ASYNC):      " : "(SYNC, press Tab): ");
	            out << std::fixed << std::setprecision(2) << wal_time << " ms (" << 1000.f / wal_time << " fps)";
	            cv::putText(curr_frame, out.str(), cv::Point2f(0, 50), cv::FONT_HERSHEY_TRIPLEX, font_size, cv::Scalar(0, 0, 255));
	            if (!isAsyncMode) {  // In the true async mode, there is no way to measure detection time directly
	                out.str("");
	                out << "Detection time  : " << std::fixed << std::setprecision(2) << det_time
	                    << " ms ("
	                    << 1000.f / det_time << " fps)";
	                cv::putText(curr_frame, out.str(), cv::Point2f(0, 75), cv::FONT_HERSHEY_TRIPLEX, font_size,
	                            cv::Scalar(255, 0, 0));
				}
                // ---------------------------Process output blobs--------------------------------------------------
                // Processing results of the CURRENT request
                const float *detections = async_infer_request_curr[(prv_idx < 0)? cur_idx : prv_idx]->GetBlob(outputName)->buffer().as<PrecisionTrait<Precision::FP32>::value_type*>();
                for (int i = 0; i < maxProposalCount; i++) {
                    float image_id = detections[i * objectSize + 0];
                    if (image_id < 0) {
                        break;
                    }

                    float confidence = detections[i * objectSize + 2];
                    auto label = static_cast<int>(detections[i * objectSize + 1]);
                    float xmin = detections[i * objectSize + 3] * width;
                    float ymin = detections[i * objectSize + 4] * height;
                    float xmax = detections[i * objectSize + 5] * width;
                    float ymax = detections[i * objectSize + 6] * height;

                    if (FLAGS_r) {
                        std::cout << "[" << i << "," << label << "] element, prob = " << confidence <<
                                  "    (" << xmin << "," << ymin << ")-(" << xmax << "," << ymax << ")"
                                  << ((confidence > FLAGS_t) ? " WILL BE RENDERED!" : "") << std::endl;
                    }

                    if (confidence > FLAGS_t) {
                        /** Drawing only objects when > confidence_threshold probability **/
                        std::ostringstream conf;
                        conf << ":" << std::fixed << std::setprecision(3) << confidence;
                        cv::putText(curr_frame,
                                    (static_cast<size_t>(label) < labels.size() ?
                                    labels[label] : std::string("label #") + std::to_string(label)) + conf.str(),
                                    cv::Point2f(xmin, ymin - 5), cv::FONT_HERSHEY_COMPLEX_SMALL, font_size,
                                    cv::Scalar(0, 0, 255));
                        cv::rectangle(curr_frame, cv::Point2f(xmin, ymin), cv::Point2f(xmax, ymax), cv::Scalar(0, 0, 255));
                    }
                }
            }
			prv_idx = cur_idx;

            if (!FLAGS_no_show) {
                cv::imshow("Detection results", curr_frame);
            }

            t1 = std::chrono::high_resolution_clock::now();
            ocv_render_time = std::chrono::duration_cast<ms>(t1 - t0).count();

            if (isLastFrame) {
                break;
            }

            if (isModeChanged) {
                isModeChanged = false;
            }

            // Final point:
            // in the truly Async mode we swap the NEXT and CURRENT requests for the next iteration
            curr_frame = next_frame;
            next_frame = cv::Mat();
            if (isAsyncMode) {
                async_infer_request_curr[cur_idx].swap(async_infer_request_next[cur_idx]);
            }

            const int key = cv::waitKey(1);
            if (27 == key)  // Esc
                break;
            if (9 == key) {  // Tab
                isAsyncMode ^= true;
                isModeChanged = true;
            } else {
                presenter.handleKey(key);
            }
        }
        // -----------------------------------------------------------------------------------------------------
        auto total_t1 = std::chrono::high_resolution_clock::now();
        ms total = std::chrono::duration_cast<ms>(total_t1 - total_t0);
        std::cout << "Total Inference time: " << total.count() << std::endl;

        /** Show performace results **/
        if (FLAGS_pc) {
		    for (int i = 0; i < vpu_count ; i ++) 
		    {
	            printPerformanceCounts(*async_infer_request_curr[i], std::cout, getFullDeviceName(ie, FLAGS_d));
			}
        }
        std::cout << presenter.reportMeans() << '\n';

    }
    catch (const std::exception& error) {
        std::cerr << "[ ERROR ] " << error.what() << std::endl;
        result = 1;
    }
    catch (...) {
        std::cerr << "[ ERROR ] Unknown/internal exception happened." << std::endl;
        result = 1;
    }

	if(result == 0)
	{
    	slog::info << "Execution successful" << slog::endl;
	}
	if(fpd_input) {
		// Disable camera input into FPGA
		if( pio_reg != NULL ) {
			if		( video_res == 1080 )	pio_reg[ALTPIO_DATA]	= 0x030e;		// 1920x1080(FHD)
			else if	( video_res == 480 )	pio_reg[ALTPIO_DATA]	= 0x033e;		//  640x480 (VGA)
		}

        slog::info << "Scaler & Frame buffer Stop" << slog::endl;
		if(scl_reg != NULL) scl_reg[ALTVIPSCL2_CONTROL] = 0x00;	// Scaler Stop
	   	if(vfb_reg != NULL) vfb_reg[ALTVIPFB2_CONTROL] = 0x00;  // Frame Buffer Stop

		// Stop Camera input
		usleep( 10000 );
		cis_stop();
		
		usleep( 100 );
		pio_reg[ALTPIO_DATA]		= 0x0000;		// Reset camera input path

		if(frame_buf != NULL) free(frame_buf);

		if(frame_mem != NULL) munmap(frame_mem, VFB_SIZE);
		if(reg_mem != NULL) munmap(reg_mem, LW_BRIDGE_SIZE);

		if(fd_vfb != 0) close(fd_vfb);
		if(fd_reg != 0) close(fd_reg);
	}
    return result;
}



